var BOARD_EMPTY = -1;
var BOARD_BLACK = 0;
var BOARD_WHITE = 1;

function BoardModel() {
  var history = new Array();
  var board;

  this.gametree = new GameNode();
  this.roots = new Array(this.gametree);
  this.currentNode = this.gametree;
  this.size = 19;

  function pushBoard() {
    var copy = new Array();

    board.forEach(function(row) {
      copy.push(row.slice(0));
    });

    copy.moveNumber = board.moveNumber;
    copy.blackPrisoners = board.blackPrisoners;
    copy.whitePrisoners = board.whitePrisoners;
    history.push(copy);
  }

  function popBoard() {
    board = history.pop();
  }

  this.initializeBoard = function() {
    board = new Array(this.size);
    
    for (var i = 0; i < this.size; i++) {
      board[i] = new Array(this.size);
    }

    board.moveNumber = 0;
    board.blackPrisoners = 0;
    board.whitePrisoners = 0;
  };

  this.objectAt = function(x, y) {
    return board[x][y];
  };

  this.getCurrentMove = function() {
    if (this.currentNode == undefined) {
      return undefined;
    }

    var move = this.currentNode.getMove();

    if (move == undefined) {
      return move;
    }

    if ((! move.pass) &&
	! ((this.size == 19) &&
	   (move.coords[0] == 19) &&
	   (move.coords[1] == 19))) {
      return move;
    } else {
      return undefined;
    }
  };

  this.getNextMove = function() {
    if (this.currentNode == undefined) {
      return undefined;
    }

    var variation = this.currentNode.variation;

    return this.currentNode.nodes[variation].getMove();
  };

  this.getMove = function(move) {
    var curMove = board.moveNumber - 1; 
    var node = this.currentNode;

    if (curMove < move) {
      for (; curMove < move; curMove++) {
        node = node.nodes[node.variation];
      }
    } else {
      for (; curMove > move; curMove--) {
        node = node.parent;
      }
    }

    return node;
  }

  this.getVariations = function(atThisMove) {
    var variations = new Array();

    if (this.currentNode == undefined) {
      return variations;
    }

    var func = function (node) {
      if (node == this.currentNode) {
        return;
      }
      
      var move = node.getMove();

      if (move != undefined) {
        variations.push(move);
      }
    };

    if (atThisMove && (this.currentNode.parent != undefined)) {
      this.currentNode.parent.nodes.forEach(func, this);
    } else if (! atThisMove && (this.currentNode.nodes.length > 1)) {
      this.currentNode.nodes.forEach(func, this);
    }

    return variations;
  };

  this.getMarkers = function() {
    var markers = new Array();

    if (this.currentNode == undefined) {
      return markers;
    }

    var node = this.currentNode;

    markers["TW"] = node.getPropertyAsCoords("TW");
    markers["TB"] = node.getPropertyAsCoords("TB");
    markers["TR"] = node.getPropertyAsCoords("TR");
    markers["SQ"] = node.getPropertyAsCoords("SQ");
    markers["CR"] = node.getPropertyAsCoords("CR");
    markers["MA"] = node.getPropertyAsCoords("MA");
    markers["SL"] = node.getPropertyAsCoords("SL");

    return markers;
  }

  this.getLabels = function() {
    var labels;
    var letters;
    var curletter = "a".charCodeAt(0);

    if (this.currentNode == undefined) {
      return new Array();
    }

    labels = this.currentNode.getPropertyAsLabel("LB");
    letters = this.currentNode.getPropertyAsCoords("L");

    if (letters != null) {
      letters.forEach(function (coords) {
                        labels.push(new Array(coords, String.fromCharCode(curletter++)));
                      });
    }

    return labels;
  }

  this.getComments = function() {
    if (this.currentNode == undefined) {
      return new Array();
    }

    return this.currentNode.getAllProperty("C");
  }

  this.getName = function() {
    if (this.currentNode == undefined) {
      return "";
    }

    return this.currentNode.getProperty("N");
  }
  
  this.getMoveNumber = function() {
    return board.moveNumber;
  }

  this.getBlackPrisoners = function() {
    return board.blackPrisoners;
  }

  this.getWhitePrisoners = function() {
    return board.whitePrisoners;
  }

  this.load = function(url) {
    var sgf = new SGF(url);
    var roots;

    try {
      this.roots = sgf.read();

      this.setGame(0);
    } catch (e) {
      alert("Error: " + e.message + " at " + e.position);
    }
  };

  this.setGame = function(index) {
    this.gametree = this.roots[index];
    this.currentNode = this.gametree;

    if (this.gametree.hasProperty("SZ")) {
      this.size = this.gametree.getProperty("SZ") * 1;
    } else {
      this.size = 19;
    }
    
    this.initializeBoard();
    this.renderCurrentNode();
  }

  this.getProperty = function(name) {
    if (this.gametree != undefined) {
      return this.gametree.getProperty(name);
    } else {
      return undefined;
    }
  }

  this.getGameNames = function() {
    var res = new Array();
    var i = 1;

    this.roots.forEach(function (g) {
      if (g.hasProperty("GN")) {
        res.push(g.getProperty("GN"));
      } else {
        res.push("Game " + i);
      }

      i++;
    });    

    return res;
  }

  this.addStone = function(x, y, colour) {
    var same = new Array();
    var different = new Array();

    var surround = new Array();

    if (x < this.size - 1) {
      surround.push(board[x + 1][y]);
    }

    if (x > 0) {
      surround.push(board[x - 1][y]);
    }

    if (y < this.size - 1) {
      surround.push(board[x][y + 1]);
    }

    if (y > 0) {
      surround.push(board[x][y - 1]);
    }

    surround.forEach(function(g) {
      if (g == undefined) {
        return;
      } else if (g.colour == colour) {
        if (same.indexOf(g) < 0) {
          same.push(g);
        }
      } else {
        if (different.indexOf(g) < 0) {
          different.push(g);
        }
      }
    });

    var group = new Group(colour);

    group.addPoint(x, y);

    same.forEach(function(g) {
      group = group.merge(g);
    });

    group.populate();

    different.forEach(function(g) {
      if (g.liberties() == 0) {
        if (g.colour == BOARD_BLACK) {
          board.blackPrisoners += g.count();
        } else if (g.colour == BOARD_WHITE) {
          board.whitePrisoners += g.count();
        }

        g.kill();
      }
    });
  }

  this.removeStone = function(x, y) {
    var group = board[x][y];

    if (group == undefined) {
      return;
    }

    group = group.clone();
    group.removePoint(x, y);

    board[x][y] = undefined;
  }

  this.up = function() {
    if (this.currentNode.parent == undefined) {
      return undefined;
    }

    this.currentNode = this.currentNode.parent;

    popBoard();

    return this.currentNode;
  }

  this.down = function() {
    if (this.currentNode.nodes.length == 0) {
      return undefined;
    }

    var next = this.currentNode.nodes[this.currentNode.variation];

    pushBoard();

    board.moveNumber++;

    this.currentNode = next;
    this.renderCurrentNode();

    return this.currentNode;
  }

  this.renderCurrentNode = function(node) {
    var node = this.currentNode;

    if (node == undefined) {
      return;
    }

    var move = node.getMove();

    if ((move != undefined) && (! move.pass) && 
	! ((this.size == 19) &&
	   (move.coords[0] == 19) &&
	   (move.coords[1] == 19))) {
      this.addStone(move.coords[0],
		    move.coords[1],
		    move.colour);
    }

    if (node.hasProperty("AW")) {
      var stones = node.getPropertyAsCoords("AW");

      stones.forEach(function(coords) {
		       this.addStone(coords[0], 
				     coords[1], 
				     BOARD_WHITE);
		     }, this);
    }

    if (node.hasProperty("AB")) {
      var stones = node.getPropertyAsCoords("AB");

      stones.forEach(function(coords) {
		       this.addStone(coords[0], 
				     coords[1], 
				     BOARD_BLACK);
		     }, this);
    }

    if (node.hasProperty("AE")) {
      var spots = node.getPropertyAsCoords("AE");

      spots.forEach(function(coords) {
		      this.removeStone(coords[0],
				       coords[1]);				       
		    }, this);
    }

    if (node.hasProperty("MN")) {
      board.moveNumber = node.getProperty("MN") * 1;
    }
  }

  this.currentMoveCoords = function() {
    if (this.currentNode == undefined) {
      return;      
    }    
  }

  this.nextVariation = function(parentVariation) {
    if (parentVariation) {
      if (this.currentNode.parent == undefined) {
	return;
      }
      
      this.up();
    }

    var curVar = this.currentNode.variation;
    
    if (curVar < this.currentNode.nodes.length - 1) {
      this.currentNode.variation++;
    }

    if (parentVariation) {
      this.down();
    }
  }

  this.prevVariation = function(parentVariation) {
    if (parentVariation) {
      if (this.currentNode.parent == undefined) {
	return;
      }
      
      this.up();
    }

    var curVar = this.currentNode.variation;

    if (curVar > 0) {
      this.currentNode.variation--;
    }

    if (parentVariation) {
      this.down();
    }
  }

  this.hasVariation = function(x, y) {
    if (this.currentNode == undefined) {
      return false;
    }

    var nodes = this.currentNode.nodes;
    var hasVar = false;
    var prop;

    nodes.forEach(function(node) {
      var move = node.getMove();

      if ((move != undefined) &&
          (move.coords[0] == x) && 
          (move.coords[1] == y)) {

        hasVar = true;
      }
    });

    return hasVar;
  }

  this.toVariation = function(x, y) {
    if (this.currentNode == undefined) {
      return;
    }

    var nodes = this.currentNode.nodes;
    var next;
    
    for (var i = 0; i < nodes.length; i++) {
      var node = nodes[i];
      var move = node.getMove();

      if ((move != undefined) &&
          (move.coords[0] == x) &&
          (move.coords[1] == y)) {

        this.currentNode.variation = i;
        break;
      }
    }

    this.down();
  }

  this.addMove = function(x, y, force) {
    var cell = board[x][y];

    if ((cell != undefined) && (! force)) {
      return;
    }

    var node = new GameNode();

    var colour;

    if (this.currentNode.hasProperty("B")) {
      colour = "W";
    } else if (this.currentNode.hasProperty("W")) {
      colour = "B";
    } else if (this.currentNode.hasProperty("PL")) {
      colour = this.currentNode.getProperty("PL");
    } else {
      colour = "B";
    }

    node.addPropertyCoords(colour, x, y);

    this.currentNode.nodes.push(node);
    this.currentNode.variation = this.currentNode.nodes.length - 1;
    node.parent = this.currentNode;

    this.down();
  }

  function Group(colour) {
    this.colour = colour;
    this.coords = new Array();

    this.addPoint = function(x, y) {
      this.coords.push(new Array(x, y));

      board[x][y] = this;
    }

    this.removePoint = function(x, y) {
      var length = this.coords.length;

      for (var i = 0; i < length; i++) {
	if ((this.coords[i][0] == x) &&
	    (this.coords[i][1] == y)) {
	  this.coords.splice(i, 1);
	  break;
	}
      }
    }

    this.merge = function(group) {
      var g = new Group(this.colour);
      
      g.coords = this.coords.concat(group.coords);

      return g;
    }

    this.liberties = function() {
      var liberties = 0;

      this.coords.forEach(function(point) {
        var x = point[0];
        var y = point[1];

        if ((x > 0) && (board[x - 1][y] == undefined)) {
          liberties++;
        }

        if ((x < board.length - 1) && (board[x + 1][y] == undefined)) {
          liberties++;
        }

        if ((y > 0) && (board[x][y - 1] == undefined)) {
          liberties++;
        }

        if ((y < board.length - 1) && (board[x][y + 1] == undefined)) {
          liberties++;
        }
      });

      return liberties;
    }

    this.populate = function() {      
      this.coords.forEach(function(p) {
        board[p[0]][p[1]] = this;
      }, this);
    }

    this.count = function() {
      return this.coords.length;
    }

    this.kill = function() {
      this.coords.forEach(function(p) {
        board[p[0]][p[1]] = undefined;
      });
    }

    this.clone = function() {
      var g = new Group(this.colour);

      g.coords = this.coords.concat(new Array());

      return g;
    }
    
    this.dump = function() {
      FoxGo.log("Group Colour: " + this.colour);
      FoxGo.log("Points: " + this.coords);
    }
  }

  this.initializeBoard();
}
